@@ -38,7 +38,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-
 #include <sys/types.h>
 #include <sys/stat.h>
 
@@ -244,7 +243,7 @@ static int _parse_cert_file(config_file *cf, FILE *fp)
     return result;
 }
 
-int keydbcfg_load_pk_file(config_file *cf)
+static int _load_pk_file(config_file *cf)
 {
     static const char pk_file_name[] = PK_FILE_NAME;
     FILE *fp;
@@ -265,7 +264,7 @@ int keydbcfg_load_pk_file(config_file *cf)
     return result;
 }
 
-int keydbcfg_load_cert_file(config_file *cf)
+static int _load_cert_file(config_file *cf)
 {
     static const char cert_file_name[] = CERT_FILE_NAME;
     FILE *fp;
@@ -402,7 +401,134 @@ int keycache_find(const char *type, const uint8_t *disc_id, uint8_t *key, unsign
     return result;
 }
 
-char *keydbcfg_find_config_file(void)
+static char *_cache_file(const char *name)
+{
+    const char *cache_dir = get_cache_home();
+
+    if (!cache_dir) {
+        return NULL;
+    }
+
+    return str_printf("%s/%s/%s", cache_dir, CFG_DIR, name);
+}
+
+int cache_save(const char *name, uint32_t version, const void *data, uint32_t len)
+{
+    int result = 0;
+    char *file = _cache_file(name);
+
+    if (file) {
+        if (_mkpath(file)) {
+            FILE *fp = fopen(file, "w");
+
+            if (fp) {
+                if (fwrite(&version, 1, 4, fp) == 4 &&
+                    fwrite(&len, 1, 4, fp) == 4 &&
+                    fwrite(data, 1, len, fp) == len) {
+                    DEBUG(DBG_FILE, "Wrote %d bytes to %s\n", len + 8, file);
+                    result = 1;
+
+                } else {
+                    DEBUG(DBG_FILE, "Error writing to %s\n", file);
+                }
+
+                fclose(fp);
+            }
+        }
+
+        X_FREE(file);
+    }
+
+    return result;
+}
+
+int cache_get(const char *name, uint32_t *version, uint32_t *len, void *buf)
+{
+    int result = 0;
+    char *file = _cache_file(name);
+
+    *version = 0;
+    if (len) {
+        *len = 0;
+    }
+
+    if (file) {
+        FILE *fp = fopen(file, "r");
+
+        if (fp) {
+            DEBUG(DBG_FILE, "Reading %s\n", file);
+
+            if (fread(version, 1, 4, fp) == 4 &&
+                (!len || fread(len, 1, 4, fp) == 4) &&
+                (!buf || fread(buf, 1, *len, fp) == *len)) {
+
+              DEBUG(DBG_FILE, "Read %d bytes from %s, version %d\n", 4 + (len ? 4 : 0) + (buf ? *len : 0), file, *version);
+              result = 1;
+
+            } else {
+              DEBUG(DBG_FILE, "Error reading from %s\n", file);
+            }
+
+            fclose(fp);
+
+        } else {
+            DEBUG(DBG_FILE, "%s not found\n", file);
+        }
+
+        X_FREE(file);
+    }
+
+    return result;
+}
+
+int cache_remove(const char *name)
+{
+    char *file = _cache_file(name);
+    int result = !remove(name);
+    if (!result) {
+        DEBUG(DBG_FILE, "Error removing %s\n", file);
+    }
+    X_FREE(file);
+    return result;
+}
+
+void *cache_get_or_update(const char *type, const void *data, uint32_t *len, uint32_t version)
+{
+    uint32_t cache_len, cache_version;
+    uint8_t *cache_data = NULL;
+
+    /* get cache version */
+    cache_get(type, &cache_version, &cache_len, NULL);
+
+    /* if cached data is later, use it */
+    if (cache_len && cache_version > version) {
+        cache_data = malloc(cache_len);
+
+        if (cache_get(type, &cache_version, &cache_len, cache_data)) {
+            DEBUG(DBG_AACS, "Using cached %s. Version: %d\n", type, cache_version);
+            *len = cache_len;
+            return cache_data;
+        }
+
+        /* read failed, fall back to older version */
+        X_FREE(cache_data);
+    }
+
+    if (data) {
+        cache_data = malloc(*len);
+        memcpy(cache_data, data, *len);
+
+        /* cached data is older, update cache */
+        if (cache_version < version) {
+            cache_save(type, version, data, *len);
+            DEBUG(DBG_AACS, "Updated cached %s. Version: %d\n", type, version);
+        }
+    }
+
+    return cache_data;
+}
+
+static char *_find_config_file(void)
 {
     static const char cfg_file_name[] = CFG_FILE_NAME;
 
@@ -422,3 +548,37 @@ char *keydbcfg_find_config_file(void)
     return cfg_file;
 }
 
+config_file *keydbcfg_config_load(const char *configfile_path)
+{
+    int config_ok = 0;
+
+    config_file *cf = keydbcfg_new_config_file();
+
+    /* try to load KEYDB.cfg */
+
+    if (configfile_path) {
+        config_ok = keydbcfg_parse_config(cf, configfile_path);
+
+    } else {
+        /* If no configfile path given, check for config files in user's home or
+         * under /etc.
+         */
+        char *cfgfile = _find_config_file();
+        config_ok = keydbcfg_parse_config(cf, cfgfile);
+        X_FREE(cfgfile);
+    }
+
+    /* Try to load simple (aacskeys) config files */
+
+    config_ok = _load_pk_file(cf)   || config_ok;
+    config_ok = _load_cert_file(cf) || config_ok;
+
+    if (!config_ok) {
+        DEBUG(DBG_AACS | DBG_CRIT, "No valid AACS configuration files found\n");
+        keydbcfg_config_file_close(cf);
+        return NULL;
+    }
+
+    return cf;
+}
+
