@@ -1,6 +1,7 @@
 /*
  * This file is part of libaacs
  * Copyright (C) 2009-2010  Obliter0n
+ * Copyright (C) 2010       npzacs
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,9 +27,16 @@
 
 typedef struct mmc MMC;
 
+#define MMC_SUCCESS             0 /* no errors */
+#define MMC_ERROR              -1 /* MMC failed */
+#define MMC_ERROR_CERT_REVOKED -2 /* revoked certificate */
+
 AACS_PRIVATE MMC *mmc_open(const char *path);
 AACS_PRIVATE void mmc_close(MMC *mmc);
 AACS_PRIVATE int  mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cert,
                                uint8_t *vid);
 
+/* read partial MKB */
+AACS_PRIVATE uint8_t *mmc_read_mkb(MMC *mmc, int address, int *size);
+
 #endif /* MMC_H_ */
